/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/testbed/socket.c 
 * @brief socket operation for communication between the testbed processes
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Krishan Ramanathan
 **/


#include "socket.h"
#include "testbed.h"

/* ********* helper methods to implement primitive commands *********** */


int sock = -1;

static void writeAll(int fd,
		     char * data,
		     unsigned int len) {
  unsigned int pos;
  int ret;
  pos = 0;
  ret = 42;
  while ( (ret > 0) &&
	  (pos < len) ) {
    ret = write(sock, &data[pos], len-pos);
    if (ret > 0) 
      pos += ret;
    else
      LOG(LOG_WARNING,
	  "WARNING: write to socket failed: %s\n",
	  strerror(errno));
  }
}

/* ************** socket operations (IPC) ************* */


typedef struct {
  unsigned int size;
  unsigned int type;
  char data[0];
} ExchangeBuffer;

void socketSend(unsigned int len,
		unsigned int type,
		void * data) {
  ExchangeBuffer * buf;
  buf = MALLOC(len + sizeof(ExchangeBuffer));
  if (len > 0)
    memcpy(&buf->data[0],
	   data, 
	   len);
  buf->size = htonl(len + sizeof(ExchangeBuffer));
  buf->type = htonl(type);
  writeAll(sock,
	   (void*)buf,
	   len+sizeof(ExchangeBuffer));
  FREE(buf);
}

/**
 * Read a message from the socket.
 * @return the type of the message
 */
unsigned int readSocket(char ** rbuf,
			unsigned int * len) {
  unsigned int type;  
  ExchangeBuffer * buf;
  unsigned int pos;
  int ret;
  unsigned int mlen;
  
  pos = 0;
  ret = 42;
  while ( (pos < sizeof(unsigned int)) &&
	  (ret > 0) ) {
    ret = read(sock,
	       &((char*)&mlen)[pos],
	       sizeof(unsigned int)-pos);
    if (ret >= 0) 
      pos += ret;
    else
      errexit("FATAL: socket protocol error (%s)\n",
	      strerror(errno));
  }
  *len = ntohl(mlen);
  buf = MALLOC(*len);
  while ( (pos < *len) &&
	  (ret > 0) ) {
    ret = read(sock,
	       &((char*)buf)[pos],
	       *len-pos);
    if (ret > 0) 
      pos += ret;
    else
      errexit("FATAL: socket protocol error (%s)\n",
	      strerror(errno));
  }  
  type = ntohl(buf->type);
  *rbuf = MALLOC((*len) - sizeof(ExchangeBuffer));
  memcpy(*rbuf,
	 &buf->data[0],
	 (*len) - sizeof(ExchangeBuffer));
  FREE(buf);
  *len -= sizeof(ExchangeBuffer);
  return type;
}
					 

void PRINTF(char * fmt,
	    ...) {  
  va_list	args;  
  unsigned int n;
  unsigned int size = 1024;
  char * p;

  p = MALLOC(size);
  while (1) {
    /* Try to print in the allocated space. */
    va_start(args, fmt);
    n = vsnprintf(&p[sizeof(unsigned int)], size, fmt, args);
    va_end(args);
    /* If that worked, return the string. */
    if (n > -1 && n < size) {
      *((int*)p) = n;
      socketSend(n, SOCKET_PRINTF, p);
      FREE(p);
      return;
    }
    /* Else try again with more space. */
    if (n > -1)    /* glibc 2.1 */
      GROW(p, size, n+1); /* precisely what is needed */
    else           /* glibc 2.0 */
      GROW(p, size, size*2);  /* twice the old size */
  }
}

/* end of socket communication primitives */
