/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Testbed CORE. This is the code that is plugged
 * into the GNUnet core to enable transport profiling.
 *
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Khrisna Ramanathan
 * @file applications/testbed/testbed.c
 **/


#include "testbed.h"

#define DEBUG_TESTBED YES

/* */
static CoreAPIForApplication * coreAPI = NULL;

/**
 * Handler that is called for "message not understood" cases.
 */
static void tb_undefined(GNUNET_TCP_SOCKET *client,
			 TESTBED_CS_MESSAGE * msg) {
  LOG(LOG_WARNING,
      "WARNING: received unknown testbed message of type %u\n",
      ntohl(msg->msgType));
}

static void tb_ADD_PEER(GNUNET_TCP_SOCKET *client,
			TESTBED_CS_MESSAGE * msg) {
  p2p_HEADER noise;
  TESTBED_ADD_PEER_MESSAGE * hm 
    = (TESTBED_ADD_PEER_MESSAGE*) msg;

  if (sizeof(TESTBED_ADD_PEER_MESSAGE) >
      ntohs(msg->header.size) ) {
    LOG(LOG_ERROR,
	"ERROR: size of ADD_PEER message is to short.  Ignoring.\n");
    return;
  }
  if (HELO_Message_size(&hm->helo) !=
      ntohs(msg->header.size) + sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_ERROR,
	"ERROR: size of ADD_PEER message is wrong.  Ignoring.\n");
    return;
  }
  coreAPI->bindAddress(&hm->helo);
  noise.size = htons(sizeof(p2p_HEADER));
  noise.requestType = htons(p2p_PROTO_NOISE);
  coreAPI->sendToNode(&hm->helo.senderIdentity,
		      &noise,
		      EXTREME_PRIORITY, /* never drop! */
		      0 /* NOW! */);
}

static void tb_DEL_PEER(GNUNET_TCP_SOCKET *client,
			TESTBED_DEL_PEER_MESSAGE * msg) {
  coreAPI->disconnectFromPeer(&msg->host);
}

static void tb_DEL_ALL_PEERS(GNUNET_TCP_SOCKET *client,
			     TESTBED_DEL_ALL_PEERS_MESSAGE * msg) {
  /* FIXME: Need to iterate over the current peers and add a 
     function call to delete the peers */
  coreAPI->disconnectPeers();
}

static void tb_GET_HELO(GNUNET_TCP_SOCKET *client,
			TESTBED_GET_HELO_MESSAGE * msg) {
  HELO_Message * helo;
  unsigned int proto = ntohs(msg->proto);
    
  if (SYSERR == coreAPI->identity2Helo(coreAPI->myIdentity, 
				       proto,
				       NO, 
				       &helo)) {
    LOG(LOG_WARNING, 
	"WARNING: TESTBED could not generate HELO message for protocol %u\n",
	proto);
    /* FIXME: well, we better report back some error code! */
  } else {
    TESTBED_ADD_PEER_MESSAGE * reply 
      = MALLOC(ntohs(helo->header.size) + sizeof(TESTBED_CS_MESSAGE));
    reply->header.header.size 
      = htons(ntohs(helo->header.size) + sizeof(TESTBED_CS_MESSAGE));
    reply->header.header.tcpType 
      = htons(TESTBED_CS_PROTO_REPLY);
    reply->header.msgType
      = htonl(TESTBED_HELO_RESPONSE);
    memcpy(&reply->helo, 
	   helo, 
	   ntohs(helo->header.size));
    writeToSocket(client,
		  &reply->header.header);
    FREE(helo);
    FREE(reply);
  }    
}

static void tb_SET_TVALUE(GNUNET_TCP_SOCKET *client,
			  TESTBED_SET_TVALUE_MESSAGE * msg) {
  /* FIXME: Need to iterate over the current peers and add a 
     function call to delete the peers */
  /* coreAPI->getTrust();*/
}			  

static void tb_GET_TVALUE(GNUNET_TCP_SOCKET *client,
			  TESTBED_GET_TVALUE_MESSAGE * msg) {
  unsigned int trust;

  trust = coreAPI->getTrust(&msg->otherPeer);
  if (OK != sendTCPResult(client,
			  (int)trust)) {
    LOG(LOG_ERROR,
	"ERROR: could not send response to GET_TVALUE back to client.\n");
  }
}			  

static void tb_BLACKLIST(GNUNET_TCP_SOCKET *client,
			 TESTBED_CS_MESSAGE * msg) {
  /* FIXME: Need to iterate over the current peers and add a 
     function call to delete the peers */
}			  

static void tb_SET_IN_BW(GNUNET_TCP_SOCKET *client,
			 TESTBED_SET_IN_BW_MESSAGE * msg) {
  /* FIXME: Need to iterate over the current peers and add a 
     function call to delete the peers */
  /*setConfigurationInt("LOAD", "MAXNETDOWNBPSTOTAL", helo->value);*/
  /* FIXME: need to add a function in util.c to change the 
     node's configuration */
}			  

static void tb_SET_OUT_BW(GNUNET_TCP_SOCKET *client,
			  TESTBED_SET_OUT_BW_MESSAGE * msg) {
  /* FIXME: Need to iterate over the current peers and add a 
     function call to delete the peers */
  /*setConfigurationInt("LOAD", "MAXNETDOWNBPSTOTAL", helo->value);*/
  /* FIXME: need to add a function in util.c to change the 
     node's configuration */
}			  

static void tb_LOAD_MODULE(GNUNET_TCP_SOCKET * client,
			   TESTBED_CS_MESSAGE * msg) {
  unsigned short size;
  char * name;
  int ok;

  size = ntohs(msg->header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid LOAD_MODULE message\n");
    return;
  } 
  name = STRNDUP(&((TESTBED_LOAD_MODULE_MESSAGE*)msg)->modulename[0],
		 size - sizeof(TESTBED_CS_MESSAGE));
  if (strlen(name) == 0) {
    LOG(LOG_WARNING,
	"WARNING: received invalid LOAD_MODULE message (empty module name)\n");
    return;
  }
  ok = coreAPI->loadApplicationModule(name);
  if (ok != OK)
    LOG(LOG_WARNING,
	"WARNING: loading module failed.  Notifying client.\n");
  FREE(name);
  if (OK != sendTCPResult(client,
			  ok)) {
    LOG(LOG_ERROR,
	"ERROR: could not notify client of LOAD_MODULE result %d\n",
	ok);
  }
}

static void tb_UNLOAD_MODULE(GNUNET_TCP_SOCKET * client,
			     TESTBED_CS_MESSAGE * msg) {
  unsigned short size;
  char * name;
  int ok;

  size = ntohs(msg->header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid UNLOAD_MODULE message\n");
    return;
  } 
  name = STRNDUP(&((TESTBED_UNLOAD_MODULE_MESSAGE*)msg)->modulename[0],
		 size - sizeof(TESTBED_CS_MESSAGE));
  if (strlen(name) == 0) {
    LOG(LOG_WARNING,
	"WARNING: received invalid UNLOAD_MODULE message (empty module name)\n");
    return;
  }
  ok = coreAPI->unloadApplicationModule(name);
  if (ok != OK)
    LOG(LOG_WARNING,
	"WARNING: unloading module failed.  Notifying client.\n");
  FREE(name);
  if (OK != sendTCPResult(client,
			  ok)) {
    LOG(LOG_ERROR,
	"ERROR: could not notify client of UNLOAD_MODULE result %d\n",
	ok);
  }
}

static void tb_SET_LOSS_RATE(GNUNET_TCP_SOCKET * client,
			     TESTBED_CS_MESSAGE * msg) {
}

/**
 * General type of a message handler.
 */			  
typedef void (*THandler)(GNUNET_TCP_SOCKET *client,
			 TESTBED_CS_MESSAGE * msg);

/**
 * @brief Entry in the handlers array that describes a testbed message handler.
 */
typedef struct HD_ {
  /**
   * function that handles these types of messages 
   **/
  THandler handler; 

  /**
   * Expected size of messages for this handler.  Checked by caller.  Use
   * 0 for variable size, in that case, the handler must check.
   */
  unsigned short expectedSize; 

  /**
   * Textual description of the handler for debugging 
   */
  char * description; 

  /**
   * The message-ID of the handler.  Used only for checking that 
   * the handler array matches the message IDs defined in testbed.h.
   * Must be equal to the index in the handler array that yields
   * this entry.
   */
  unsigned int msgId; 
} HD;

/* some macros to make initializing the handlers array extremely brief. */

#define TBDENTRY(a)  {           &tb_##a,             0                , "##a##", TESTBED_##a }
#define TBSENTRY(a)  { (THandler)&tb_##a, sizeof(TESTBED_##a##_MESSAGE), "##a##", TESTBED_##a }

/**
 * The array of message handlers.  Add new handlers here.
 */
static HD handlers[] = {
  TBSENTRY(undefined),  /* use "undefined" for IDs that we should never receive */
  TBDENTRY(ADD_PEER), 
  TBSENTRY(DEL_PEER), 
  TBSENTRY(DEL_ALL_PEERS),
  TBSENTRY(GET_HELO), 
  TBSENTRY(SET_TVALUE), 
  TBSENTRY(GET_TVALUE), 
  TBSENTRY(BLACKLIST), 
  TBSENTRY(SET_IN_BW), 
  TBSENTRY(SET_OUT_BW),
  TBSENTRY(SET_LOSS_RATE),
  TBDENTRY(LOAD_MODULE),
  TBDENTRY(UNLOAD_MODULE),
  { NULL, 0, NULL, 0 }, /* this entry is used to ensure that
			   a wrong TESTBED_MAX_MSG will abort
			   insted of possibly segfaulting.
			   This must always be the LAST entry. */
};

/**
 * Number of entries in handlers array.  Checked automatically when the
 * module is initialized.
 */
#define TESTBED_MAX_MSG 13


/**
 * Global handler called by the GNUnet core.  Does the demultiplexing
 * on the testbed-message type.
 */
static void csHandleTestbedRequest(GNUNET_TCP_SOCKET *client,
				   CS_HEADER * message) {
  TESTBED_CS_MESSAGE * msg;
  unsigned short size;
  unsigned int id;

#if DEBUG_TESTBED
  LOG(LOG_DEBUG, 
      "DEBUG: TESTBED handleTestbedRequest\n");
#endif
  size = ntohs(message->size);
  if (size < sizeof(TESTBED_CS_MESSAGE)) {
    LOG(LOG_WARNING,
	"WARNING: received invalid testbed message of size %u\n",
	size);
    return;
  }
  msg = (TESTBED_CS_MESSAGE *)message;
  msg->msgType = ntohl(msg->msgType);
  id = ntohl(msg->msgType);
  if (id < TESTBED_MAX_MSG) {
    if ( (handlers[id].expectedSize == 0) ||
	 (handlers[id].expectedSize == size) ) {
      LOG(LOG_DEBUG, 
	  "DEBUG: TESTBED received %s message\n",
	  handlers[id].description);
      handlers[id].handler(client, msg);
    } else {
      LOG(LOG_ERROR,
	  "ERROR: received message of type %u but unexpected size %u, expected %u\n",
	  id, 
	  size,
	  handlers[id].expectedSize);
    }
  } else {
    tb_undefined(client, msg);
  }
}


/**
 * Initialize the TESTBED module. This method name must match
 * the library name (libgnunet_XXX => initialize_XXX).
 * @return SYSERR on errors
 **/
int initialize_testbed_protocol(CoreAPIForApplication * capi) {
  int i;
  int ok = OK;

  /* some checks */
  for (i=0;i<TESTBED_MAX_MSG;i++)
    if ( (handlers[i].msgId != i) &&
	 (handlers[i].handler != &tb_undefined) )
      errexit("FATAL: Assertion failed: "
	      "Malformed handlers array in testbed.c. "
	      "Aborting.\n");
  if (handlers[TESTBED_MAX_MSG].handler != NULL)
    errexit("FATAL: Assertion failed: "
	    "TESTBED_MAX_MSG in testbed.c is wrong."
	    "Aborting.\n");

  LOG(LOG_DEBUG,
      "DEBUG: TESTBED registering handler %d\n",
      TESTBED_CS_PROTO_REQUEST);
  coreAPI = capi;
  if (SYSERR == capi->registerClientHandler(TESTBED_CS_PROTO_REQUEST,
					    (CSHandler)&csHandleTestbedRequest))
    ok = SYSERR;
  return ok;
}

void done_testbed_protocol() {
  LOG(LOG_DEBUG,
      "DEBUG: TESTBED unregistering handler %d\n",
      TESTBED_CS_PROTO_REQUEST);
  coreAPI->unregisterClientHandler(TESTBED_CS_PROTO_REQUEST,
				   (CSHandler)&csHandleTestbedRequest);
  coreAPI = NULL;
}

/* end of testbed.c */
